from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, TypeVar, cast, get_args, get_origin, get_type_hints

if TYPE_CHECKING:
    from collections.abc import Callable

from spritze.entities.provider import Provider
from spritze.entities.transient import Transient
from spritze.infrastructure.exceptions import InvalidProvider
from spritze.value_objects.scope import Scope

T = TypeVar("T")


class ProviderRegistryService:
    @staticmethod
    def register_function_providers(
        container_instance: object, providers: dict[type[object], Provider]
    ) -> None:
        func_members: list[tuple[str, object]] = list(
            inspect.getmembers(
                container_instance.__class__, predicate=inspect.isfunction
            )
        )
        for name, func_obj in func_members:
            if hasattr(func_obj, "__spritze_provider__"):
                ProviderRegistryService._register_single_function_provider(
                    container_instance, name, func_obj, providers
                )

    @staticmethod
    def register_transient_providers(
        container_instance: object, providers: dict[type[object], Provider]
    ) -> None:
        class_items: list[tuple[str, object]] = list(
            vars(container_instance.__class__).items()
        )
        for _name, attr in class_items:
            if isinstance(attr, Transient):
                ProviderRegistryService._register_single_transient_provider(
                    attr, providers
                )

    @staticmethod
    def _register_single_function_provider(
        container_instance: object,
        name: str,
        func_obj: object,
        providers: dict[type[object], Provider],
    ) -> None:
        meta_raw = getattr(func_obj, "__spritze_provider__", None)
        if meta_raw is None:
            return
        meta: dict[str, object] = cast("dict[str, object]", meta_raw)
        scope_val = meta.get("scope")
        if not isinstance(scope_val, Scope):
            raise InvalidProvider("Invalid scope on provider")

        ret_type = ProviderRegistryService._extract_return_type(func_obj)
        if ret_type is None:
            raise InvalidProvider(
                "Provider must declare a concrete return type annotation"
            )

        bound_method = cast("Callable[..., object]", getattr(container_instance, name))
        providers[ret_type] = Provider(
            func=bound_method, scope=scope_val, return_type=ret_type
        )

    @staticmethod
    def _register_single_transient_provider(
        transient_attr: Transient, providers: dict[type[object], Provider]
    ) -> None:
        target = transient_attr.target
        ann_map_ctor = cast(
            "dict[str, object]",
            get_type_hints(target.__init__, include_extras=False),
        )

        def ctor_provider(**kwargs: object) -> object:
            return target(**kwargs)

        func_annotations: dict[str, object] = {}
        for pname, ann_obj in ann_map_ctor.items():
            if pname == "self" or pname == "return":
                continue
            if isinstance(ann_obj, type):
                func_annotations[pname] = ann_obj
        func_annotations["return"] = target
        ctor_provider.__annotations__ = func_annotations

        providers[target] = Provider(
            func=cast("Callable[..., object]", ctor_provider),
            scope=Scope.REQUEST,
            return_type=target,
        )

    @staticmethod
    def _extract_return_type(func_obj: object) -> type[object] | None:
        ann_map: dict[str, object] = get_type_hints(func_obj, include_extras=False)
        ret_obj = ann_map.get("return")

        if isinstance(ret_obj, type):
            return cast("type[object]", ret_obj)

        origin_obj = get_origin(ret_obj)
        origin_name = getattr(origin_obj, "__qualname__", "")
        if origin_name in ("Generator", "AsyncGenerator"):
            args = get_args(ret_obj)
            if args and isinstance(args[0], type):
                return cast("type[object]", args[0])

        return None


__all__ = ["ProviderRegistryService"]



from __future__ import annotations

from collections.abc import Callable
from typing import Final, ParamSpec, TypeVar

from spritze.entities.transient import Transient
from spritze.value_objects.scope import Scope

P = ParamSpec("P")
R = TypeVar("R")

PROVIDER_TAG: Final[str] = "__spritze_provider__"


def provider(
    *, scope: Scope = Scope.REQUEST
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    def decorator(func: Callable[P, R]) -> Callable[P, R]:
        setattr(func, PROVIDER_TAG, {"scope": scope})
        return func

    return decorator


def singleton(func: Callable[P, R]) -> Callable[P, R]:
    return provider(scope=Scope.APP)(func)


def transient(t: type[object]) -> object:
    return Transient(t)


__all__ = ["PROVIDER_TAG", "provider", "singleton", "transient"]



from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Generic, TypeVar, override

T = TypeVar("T")


class DependencyMarker(Generic[T]):
    def __init__(self, dependency_type: type[T] | None = None) -> None:
        self.dependency_type: type[T] | None = dependency_type


DependsMarker = DependencyMarker


if TYPE_CHECKING:

    class _DependsMeta(type):
        def __class_getitem__(
            cls,
            item: type[T],
        ) -> Annotated[type[T], DependencyMarker[T]]: ...

        @override
        def __call__(
            cls,
            dependency_type: type[T] | None = None,
        ) -> DependencyMarker[T]: ...

        def __getitem__(
            cls, item: type[T]
        ) -> Annotated[type[T], DependencyMarker[T]]: ...

else:

    def _create_depends_annotation(
        t: type[T],
    ) -> Annotated[type[T], DependencyMarker[T]]:
        return Annotated[t, DependencyMarker(t)]

    class _DependsMeta(type):
        def __class_getitem__(
            cls,
            item: type[T],
        ) -> type[T]:
            return item

        def __call__(
            cls,
            dependency_type: type[T] | None = None,
        ) -> DependencyMarker[T]:
            return DependencyMarker(dependency_type)

        def __getitem__(cls, item: type[T]) -> Annotated[type[T], DependencyMarker[T]]:
            return _create_depends_annotation(item)


class Depends(Generic[T], metaclass=_DependsMeta):
    pass


__all__ = ["DependencyMarker", "Depends", "DependsMarker"]



import pytest

from spritze import provider
from spritze.infrastructure.exceptions import (
    CyclicDependency,
    DependencyNotFound,
    InvalidProvider,
)
from spritze.repositories.container_repository import Container


class A:
    pass


class B:
    pass


def test_missing_provider():
    class Empty(Container):
        pass

    c = Empty()
    with pytest.raises(DependencyNotFound):
        _ = c.resolve(str)


def test_provider_exception():
    class Bad(Container):
        @provider()
        def fail(self) -> int:
            raise ValueError("fail")

    c = Bad()
    with pytest.raises(ValueError):
        _ = c.resolve(int)


def test_cyclic_dependency():
    class Cyclic(Container):
        @provider()
        def a(self, b: B) -> A:
            assert isinstance(b, B)
            return A()

        @provider()
        def b(self, a: A) -> B:
            assert isinstance(a, A)
            return B()

    c = Cyclic()

    with pytest.raises(CyclicDependency):
        _ = c.resolve(A)


def test_provider_missing_return_annotation():
    class MissingAnnotationContainer(Container):
        @provider()
        def no_return_type(self):
            return 1

    with pytest.raises(InvalidProvider):
        _ = MissingAnnotationContainer()



from __future__ import annotations

from typing import Generic, Protocol, TypeVar, cast

T = TypeVar("T")


class ContextProvider(Protocol):
    def get_context_value(self, t: type[object]) -> object | None: ...


class ContextField(Generic[T]):
    def __init__(self, ctx_type: type[T]) -> None:
        self.ctx_type: type[T] = ctx_type
        self.name: str | None = None

    def __set_name__(self, owner: type[object], name: str) -> None:
        self.name = name

    def __get__(
        self, instance: ContextProvider | None, owner: type[object]
    ) -> ContextField[T] | T:
        if instance is None:
            return self
        value = instance.get_context_value(self.ctx_type)
        if value is None:
            raise LookupError(f"Context value for {self.ctx_type.__name__} not found")
        return cast("T", value)


class _ContextAccessor:
    def get(self, t: type[T]) -> ContextField[T]:
        return ContextField(t)


context = _ContextAccessor()

__all__ = ["ContextField", "context"]



from __future__ import annotations


class SpritzeError(Exception):
    pass


class DependencyNotFound(SpritzeError):
    dependency_type: type[object]

    def __init__(self, dependency_type: type[object]) -> None:
        super().__init__(
            f"Dependency of type '{dependency_type.__name__}' not found. "
            + "Make sure it's registered as a provider or transient dependency."
        )
        self.dependency_type = dependency_type


class InvalidProvider(SpritzeError):
    def __init__(self, message: str) -> None:
        super().__init__(f"Invalid provider configuration: {message}")


class CyclicDependency(SpritzeError):
    stack: tuple[type[object], ...]

    def __init__(self, stack: tuple[type[object], ...]) -> None:
        path = " -> ".join(t.__name__ for t in stack)
        super().__init__(
            f"Cyclic dependency detected: {path}. "
            + "Review your dependency graph to break the cycle."
        )
        self.stack = stack


class AsyncSyncMismatch(SpritzeError):
    def __init__(self, dependency_type: type[object], context: str) -> None:
        super().__init__(
            f"Cannot resolve async provider for '{dependency_type.__name__}' "
            + f"in {context} context. Use resolve_async() for async dependencies."
        )


__all__ = [
    "AsyncSyncMismatch",
    "CyclicDependency",
    "DependencyNotFound",
    "InvalidProvider",
    "SpritzeError",
]



from __future__ import annotations

from typing import TypeVar

_C = TypeVar("_C")


class Transient:
    def __init__(self, target: type[object]) -> None:
        self.target: type[object] = target
        self.attr_name: str | None = None

    def __set_name__(self, owner: type[_C], name: str) -> None:
        self.attr_name = name

    def __get__(
        self,
        instance: _C | None,
        owner: type[_C],
    ) -> type[object]:
        return self.target


__all__ = ["Transient"]



from typing import Annotated

import pytest

from spritze import Container, Depends, init, inject, singleton, transient
from spritze.infrastructure.context import ContextField, context
from spritze.infrastructure.exceptions import InvalidProvider


class Config:
    def __init__(self, db_dsn: str = "sqlite+aiosqlite:///:memory:") -> None:
        self.db_dsn: str = db_dsn


class EngineProtocol:
    def __init__(self, dsn: str) -> None:
        self.dsn: str = dsn


class AsyncEngine(EngineProtocol):
    pass


class AsyncSession:
    def __init__(self, engine: EngineProtocol) -> None:
        self.engine: EngineProtocol = engine


class AppContainer(Container):
    config: ContextField[Config] = context.get(Config)

    @singleton
    async def db_engine(self, config: Config) -> EngineProtocol:
        return AsyncEngine(config.db_dsn)

    async_session: object = transient(AsyncSession)


@pytest.mark.asyncio
async def test_integration_async():
    container = AppContainer()
    container.context.update(Config=Config(db_dsn="memory"))

    init(container)

    @inject
    async def handler(
        s: Annotated[AsyncSession, Depends()],
        e: Annotated[EngineProtocol, Depends()],
    ) -> tuple[str, str]:
        return s.engine.dsn, e.dsn

    dsn1, dsn2 = await handler()
    assert dsn1 == "memory"
    assert dsn2 == "memory"


def test_sync_inject_rejects_async_provider():
    container = AppContainer()
    container.context.update(Config=Config())

    init(container)

    @inject
    def f(e: Annotated[EngineProtocol, Depends()]) -> None:
        assert isinstance(e, AsyncEngine)
        return None

    with pytest.raises(InvalidProvider):
        f()



from collections.abc import AsyncGenerator, Generator

import pytest

from spritze import Container, Depends, init, inject, provider


def test_sync_generator_provider():
    class Gen(Container):
        @provider()
        def res(self) -> Generator[str, None, None]:
            yield "ok"

    c = Gen()
    init(c)

    @inject
    def handler(res: Depends[str]):
        return res

    assert handler() == "ok"


@pytest.mark.asyncio
async def test_async_generator_provider():
    class Gen(Container):
        @provider()
        async def res(self) -> AsyncGenerator[str, None]:
            yield "async_ok"

    c = Gen()
    init(c)

    @inject
    async def handler(res: Depends[str]):
        return res

    assert await handler() == "async_ok"



from __future__ import annotations

import inspect
from contextlib import asynccontextmanager, contextmanager
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable

    from ..value_objects.scope import Scope

from ..value_objects.provider_type import ProviderType


def _determine_provider_type(
    func: Callable[..., object],
) -> tuple[ProviderType, Callable[..., object]]:
    if inspect.isasyncgenfunction(func):
        return ProviderType.ASYNC_GEN, asynccontextmanager(func)
    elif inspect.isgeneratorfunction(func):
        return ProviderType.GEN, contextmanager(func)
    elif inspect.iscoroutinefunction(func):
        return ProviderType.ASYNC, func
    else:
        return ProviderType.SIMPLE, func


@dataclass(kw_only=True)
class Provider:
    func: Callable[..., object]
    scope: Scope
    return_type: type[object]
    provider_type: ProviderType = field(init=False)

    def __post_init__(self) -> None:
        self.provider_type, self.func = _determine_provider_type(self.func)

    @property
    def is_context_manager(self) -> bool:
        return self.provider_type in (ProviderType.GEN, ProviderType.ASYNC_GEN)

    @property
    def is_async(self) -> bool:
        return self.provider_type in (ProviderType.ASYNC, ProviderType.ASYNC_GEN)


__all__ = ["Provider"]



from spritze import Container, Depends, Scope, init, inject, provider


class ServiceA:
    pass


class ServiceB:
    def __init__(self, a: ServiceA):
        self.a: ServiceA = a


class ServiceC:
    def __init__(self, b: ServiceB):
        self.b: ServiceB = b


def test_sync_injection():
    class SyncContainer(Container):
        @provider(scope=Scope.APP)
        def x(self) -> int:
            return 42

        @provider(scope=Scope.REQUEST)
        def y(self, x: int) -> str:
            return f"Y{x}"

    c = SyncContainer()
    init(c)

    @inject
    def handler(y: Depends[str], x: Depends[int]):
        return y, x

    y, x = handler()
    assert y == "Y42"
    assert x == 42


def test_deep_dependency():
    class TestContainer(Container):
        __test__: bool = False

        @provider(scope=Scope.APP)
        def a(self) -> ServiceA:
            return ServiceA()

        @provider(scope=Scope.REQUEST)
        def b(self, a: ServiceA) -> ServiceB:
            return ServiceB(a=a)

        @provider(scope=Scope.REQUEST)
        def c(self, b: ServiceB) -> ServiceC:
            return ServiceC(b=b)

    c = TestContainer()
    init(c)

    @inject
    def handler(c: Depends[ServiceC]):
        return c

    result = handler()
    assert isinstance(result, ServiceC)
    assert isinstance(result.b, ServiceB)
    assert isinstance(result.b.a, ServiceA)



from __future__ import annotations

import inspect
from collections.abc import AsyncGenerator, Awaitable, Callable, Generator
from contextlib import (
    AbstractAsyncContextManager,
    AbstractContextManager,
    AsyncExitStack,
    ExitStack,
)
from contextvars import ContextVar
from functools import wraps
from types import MappingProxyType
from typing import (
    ParamSpec,
    TypeVar,
    cast,
    get_args,
    get_origin,
    get_type_hints,
)

from spritze.entities.provider import Provider
from spritze.entities.transient import Transient
from spritze.infrastructure.context import ContextField
from spritze.infrastructure.exceptions import (
    CyclicDependency,
    DependencyNotFound,
    InvalidProvider,
)
from spritze.services.resolution_service import ResolutionService
from spritze.value_objects.dependency_marker import DependsMarker
from spritze.value_objects.provider_type import ProviderType
from spritze.value_objects.scope import Scope

T = TypeVar("T")
P = ParamSpec("P")
R = TypeVar("R")
TypeMap = dict[str, type[object]]


class _BoundContext:
    def __init__(self, container: "Container") -> None:
        self._container: Container = container

    def update(self, **kwargs: object) -> None:
        declared: TypeMap = {}
        class_attrs: MappingProxyType[str, object] = vars(self._container.__class__)
        for name, attr in class_attrs.items():
            if isinstance(attr, ContextField):
                cf = cast("ContextField[object]", attr)
                declared[name] = cf.ctx_type
                declared[cf.ctx_type.__name__] = cf.ctx_type

        for key, value in kwargs.items():
            t = declared.get(key)
            if t is None:
                vt = type(value)
                if vt.__name__ in declared:
                    t = declared[vt.__name__]
            if t is None:
                raise InvalidProvider(f"Unknown context key: {key}")
            self._container.set_context_value(t, value)


class Container:
    def __init__(self) -> None:
        self._providers: dict[type[object], Provider] = {}
        self._app_scoped_instances: dict[type[object], object] = {}
        self._request_scoped_instances: ContextVar[
            dict[type[object], object] | None
        ] = ContextVar("spritze_request_instances", default=None)
        self._async_exit_stack: ContextVar[AsyncExitStack] = ContextVar(
            "spritze_async_exit_stack"
        )
        self._sync_exit_stack: ContextVar[ExitStack] = ContextVar(
            "spritze_sync_exit_stack"
        )
        self._context_values: dict[type[object], object] = {}
        self._resolution_stack: ContextVar[tuple[type[object], ...]] = ContextVar(
            "spritze_resolution_stack", default=()
        )

        self._register_providers()

    @property
    def context(self) -> _BoundContext:
        return _BoundContext(self)

    def get_context_value(self, t: type[object]) -> object | None:
        return self._context_values.get(t)

    def set_context_value(self, t: type[object], value: object) -> None:
        self._context_values[t] = value

    def _register_providers(self) -> None:
        self._register_function_providers()
        self._register_transient_providers()

    def _register_function_providers(self) -> None:
        for name, func_obj in inspect.getmembers(
            self.__class__, predicate=inspect.isfunction
        ):
            if hasattr(func_obj, "__spritze_provider__"):
                self._register_single_function_provider(name, func_obj)

    def _register_single_function_provider(self, name: str, func_obj: object) -> None:
        meta: dict[str, object] = cast(
            "dict[str, object]", getattr(func_obj, "__spritze_provider__")
        )
        scope_val = meta.get("scope")
        if not isinstance(scope_val, Scope):
            raise InvalidProvider("Invalid scope on provider")

        ret_type = self._extract_return_type(func_obj)
        if ret_type is None:
            raise InvalidProvider(
                "Provider must declare a concrete return type annotation"
            )

        bound_method = cast("Callable[..., object]", getattr(self, name))
        self._providers[ret_type] = Provider(
            func=bound_method, scope=scope_val, return_type=ret_type
        )

    def _extract_return_type(self, func_obj: object) -> type[object] | None:
        ann_map: dict[str, object] = get_type_hints(func_obj, include_extras=False)
        ret_obj = ann_map.get("return")

        if isinstance(ret_obj, type):
            return cast("type[object]", ret_obj)

        origin_obj = get_origin(ret_obj)
        origin_name = getattr(origin_obj, "__qualname__", "")
        if origin_name in (
            Generator.__qualname__,
            AsyncGenerator.__qualname__,
        ):
            args = get_args(ret_obj)
            if args and isinstance(args[0], type):
                return cast("type[object]", args[0])

        return None

    def _register_transient_providers(self) -> None:
        class_vars: MappingProxyType[str, object] = vars(self.__class__)
        for _name, attr in class_vars.items():
            if isinstance(attr, Transient):
                self._register_single_transient_provider(attr)

    def _register_single_transient_provider(self, transient_attr: Transient) -> None:
        target = transient_attr.target
        ann_map_ctor = cast(
            "dict[str, object]",
            get_type_hints(target.__init__, include_extras=False),
        )

        def ctor_provider(**kwargs: object) -> object:
            return target(**kwargs)

        func_annotations: dict[str, object] = {}
        for pname, ann_obj in ann_map_ctor.items():
            if pname == "self" or pname == "return":
                continue
            if isinstance(ann_obj, type):
                func_annotations[pname] = ann_obj
        func_annotations["return"] = target
        ctor_provider.__annotations__ = func_annotations

        self._providers[target] = Provider(
            func=cast("Callable[..., object]", ctor_provider),
            scope=Scope.REQUEST,
            return_type=target,
        )

    def _check_cache(self, dependency_type: type[T]) -> T | None:
        if dependency_type in self._app_scoped_instances:
            return cast("T", self._app_scoped_instances[dependency_type])
        if dependency_type in self._context_values:
            return cast("T", self._context_values[dependency_type])
        request_cache = self._request_scoped_instances.get()
        if request_cache is not None and dependency_type in request_cache:
            return cast("T", request_cache[dependency_type])
        return None

    def _cache_instance(
        self, dependency_type: type[object], instance: object, scope: Scope
    ) -> None:
        if scope == Scope.APP:
            self._app_scoped_instances[dependency_type] = instance
        elif scope == Scope.REQUEST:
            request_cache: dict[type[object], object] | None = (
                self._request_scoped_instances.get()
            )
            if request_cache is None:
                request_cache = {}
                _ = self._request_scoped_instances.set(request_cache)
            request_cache[dependency_type] = instance

    def _resolve_provider_dependencies_sync(
        self, provider: Provider
    ) -> dict[str, object]:
        kwargs: dict[str, object] = {}
        deps = ResolutionService.get_deps_to_resolve(provider.func)
        for name, dep_t in deps.items():
            kwargs[name] = self.resolve(dep_t)
        return kwargs

    async def _resolve_provider_dependencies_async(
        self, provider: Provider
    ) -> dict[str, object]:
        kwargs: dict[str, object] = {}
        deps = ResolutionService.get_deps_to_resolve(provider.func)
        for name, dep_t in deps.items():
            kwargs[name] = await self.resolve_async(dep_t)
        return kwargs

    def _execute_provider(
        self, provider: Provider, kwargs: dict[str, object]
    ) -> object:
        if provider.is_context_manager:
            cm_raw = provider.func(**kwargs)
            cm = cast("AbstractContextManager[object]", cm_raw)
            return self._sync_exit_stack.get().enter_context(cm)
        else:
            return provider.func(**kwargs)

    async def _execute_provider_async(
        self, provider: Provider, kwargs: dict[str, object]
    ) -> object:
        if provider.is_context_manager:
            exit_stack = self._async_exit_stack.get()
            if provider.provider_type == ProviderType.ASYNC_GEN:
                acm_raw = provider.func(**kwargs)
                acm = cast("AbstractAsyncContextManager[object]", acm_raw)
                return await exit_stack.enter_async_context(acm)
            else:
                cm_raw = provider.func(**kwargs)
                cm = cast("AbstractContextManager[object]", cm_raw)
                return exit_stack.enter_context(cm)
        elif provider.provider_type == ProviderType.ASYNC:
            coro = cast("Callable[..., Awaitable[object]]", provider.func)
            return await coro(**kwargs)
        else:
            return provider.func(**kwargs)

    def _resolve_with_cycle_check_sync(self, dependency_type: type[T]) -> T:
        cached = self._check_cache(dependency_type)
        if cached is not None:
            return cached

        stack = self._resolution_stack.get()
        if dependency_type in stack:
            raise CyclicDependency(stack + (dependency_type,))

        token_r = self._resolution_stack.set(stack + (dependency_type,))
        try:
            provider = self._providers.get(dependency_type)
            if provider is None:
                raise DependencyNotFound(dependency_type)

            if provider.is_async:
                raise InvalidProvider(
                    "Cannot resolve async provider for "
                    + f"{dependency_type.__name__} in sync context"
                )

            kwargs = self._resolve_provider_dependencies_sync(provider)
            instance_obj = self._execute_provider(provider, kwargs)

            self._cache_instance(dependency_type, instance_obj, provider.scope)
            return cast("T", instance_obj)
        finally:
            self._resolution_stack.reset(token_r)

    async def _resolve_with_cycle_check_async(self, dependency_type: type[T]) -> T:
        cached = self._check_cache(dependency_type)
        if cached is not None:
            return cached

        stack = self._resolution_stack.get()
        if dependency_type in stack:
            raise CyclicDependency(stack + (dependency_type,))

        token_r = self._resolution_stack.set(stack + (dependency_type,))
        try:
            provider = self._providers.get(dependency_type)
            if provider is None:
                raise DependencyNotFound(dependency_type)

            kwargs = await self._resolve_provider_dependencies_async(provider)
            instance_obj = await self._execute_provider_async(provider, kwargs)

            self._cache_instance(dependency_type, instance_obj, provider.scope)
            return cast("T", instance_obj)
        finally:
            self._resolution_stack.reset(token_r)

    def resolve(self, dependency_type: type[T]) -> T:
        return self._resolve_with_cycle_check_sync(dependency_type)

    async def resolve_async(self, dependency_type: type[T]) -> T:
        return await self._resolve_with_cycle_check_async(dependency_type)

    def injector(self) -> Callable[[Callable[P, R]], Callable[..., R]]:
        def decorator(func: Callable[P, R]) -> Callable[..., R]:
            sig = inspect.signature(func)
            ann_map = get_type_hints(func, include_extras=True)
            deps = ResolutionService.extract_dependencies_from_signature(sig, ann_map)

            def _inject_dependencies_sync(bound: inspect.BoundArguments) -> None:
                for name, t in deps.items():
                    needs_inject = name not in bound.arguments or isinstance(
                        bound.arguments.get(name), DependsMarker
                    )
                    if needs_inject:
                        bound.arguments[name] = self.resolve(t)

            async def _inject_dependencies_async(bound: inspect.BoundArguments) -> None:
                for name, t in deps.items():
                    needs_inject = name not in bound.arguments or isinstance(
                        bound.arguments.get(name), DependsMarker
                    )
                    if needs_inject:
                        bound.arguments[name] = await self.resolve_async(t)

            if inspect.iscoroutinefunction(func):

                @wraps(func)
                async def _awrapper(*args: object, **kwargs: object) -> object:
                    async with AsyncExitStack() as stack:
                        token_s = self._async_exit_stack.set(stack)
                        token_c = self._request_scoped_instances.set({})
                        try:
                            bound = sig.bind_partial(*args, **kwargs)
                            await _inject_dependencies_async(bound)
                            coro = cast("Callable[..., Awaitable[object]]", func)
                            result: object = await coro(**bound.arguments)
                            return result
                        finally:
                            self._async_exit_stack.reset(token_s)
                            self._request_scoped_instances.reset(token_c)

                from contextlib import suppress

                with suppress(Exception):
                    setattr(_awrapper, "__signature__", sig.replace(parameters=()))
                return cast("Callable[..., R]", _awrapper)
            else:

                @wraps(func)
                def _swrapper(*args: object, **kwargs: object) -> object:
                    with ExitStack() as stack:
                        token_s = self._sync_exit_stack.set(stack)
                        token_c = self._request_scoped_instances.set({})
                        try:
                            bound = sig.bind_partial(*args, **kwargs)
                            _inject_dependencies_sync(bound)
                            f_callable = cast("Callable[..., object]", func)
                            return f_callable(**bound.arguments)
                        finally:
                            self._sync_exit_stack.reset(token_s)
                            self._request_scoped_instances.reset(token_c)

                from contextlib import suppress

                with suppress(Exception):
                    setattr(_swrapper, "__signature__", sig.replace(parameters=()))
                return cast("Callable[..., R]", _swrapper)

        return cast("Callable[[Callable[P, R]], Callable[..., R]]", decorator)


__all__ = ["Container", "Scope", "Transient", "Provider"]



from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar, cast, get_args, get_type_hints

if TYPE_CHECKING:
    import inspect
    from collections.abc import Callable

T = TypeVar("T")
TypeMap = dict[str, type[object]]


class ResolutionService:
    @staticmethod
    def get_deps_to_resolve(func: Callable[..., object]) -> TypeMap:
        deps: dict[str, type[object]] = {}
        ann_map = cast("dict[str, object]", get_type_hints(func, include_extras=False))
        for name, ann_obj in ann_map.items():
            if name in ("self", "return"):
                continue
            if isinstance(ann_obj, type):
                deps[name] = ann_obj
        return deps

    @staticmethod
    def extract_dependencies_from_signature(
        sig: inspect.Signature, ann_map: dict[str, object]
    ) -> TypeMap:
        deps: dict[str, type[object]] = {}

        for p in sig.parameters.values():
            ann_obj: object | None = ann_map.get(p.name)
            dep_type = ResolutionService._extract_dependency_type(p, ann_obj)
            if dep_type is not None:
                deps[p.name] = dep_type

        return deps

    @staticmethod
    def _extract_dependency_type(
        param: inspect.Parameter, ann_obj: object | None
    ) -> type[object] | None:
        from spritze.value_objects.dependency_marker import Depends, DependsMarker

        if isinstance(cast("object", param.default), DependsMarker):
            dm_def = cast("DependsMarker[object]", param.default)
            if isinstance(ann_obj, type):
                return ann_obj
            if isinstance(dm_def.dependency_type, type):
                return dm_def.dependency_type
            return None

        if getattr(ann_obj, "__origin__", None) is Depends:
            args = cast("tuple[object, ...]", get_args(ann_obj))
            if args and isinstance(args[0], type):
                return cast("type[object]", args[0])

        args_tuple = cast("tuple[object, ...]", get_args(ann_obj))
        if args_tuple and len(args_tuple) >= 2:
            base = args_tuple[0]
            meta = args_tuple[1:]
            dep_markers: list[DependsMarker[object]] = [
                cast("DependsMarker[object]", m)
                for m in meta
                if isinstance(m, DependsMarker)
            ]
            if dep_markers:
                dm = dep_markers[0]
                if isinstance(dm.dependency_type, type):
                    return dm.dependency_type
                if isinstance(base, type):
                    return base

        return None


__all__ = ["ResolutionService"]



from .context import ContextField, context
from .exceptions import (
    CyclicDependency,
    DependencyNotFound,
    InvalidProvider,
    SpritzeError,
)

__all__ = [
    "ContextField",
    "CyclicDependency",
    "DependencyNotFound",
    "InvalidProvider",
    "SpritzeError",
    "context",
]



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, ParamSpec, TypeVar, cast
from weakref import WeakKeyDictionary

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable, Sequence

from spritze.infrastructure.exceptions import DependencyNotFound, InvalidProvider
from spritze.repositories.container_repository import Container

P = ParamSpec("P")
R = TypeVar("R")

_default_container: Container | Sequence[Container] | None = None
_WRAPPER_CACHE: WeakKeyDictionary[
    Container, dict[Callable[..., object], Callable[..., object]]
] = WeakKeyDictionary()


def init(container: Container | Sequence[Container]) -> None:
    global _default_container

    if isinstance(container, (list, tuple)) and not container:
        raise ValueError("Container sequence cannot be empty")

    _default_container = container


def _get_container() -> Container | Sequence[Container]:
    container = _default_container
    if container is None:
        raise RuntimeError(
            "No global container is set. Call spritze.init(container) first."
        )
    return container


def _get_inner(
    container: Container, func: Callable[..., object]
) -> Callable[..., object]:
    mapping = _WRAPPER_CACHE.get(container)
    if mapping is None:
        mapping = {}
        _WRAPPER_CACHE[container] = mapping
    inner = mapping.get(func)
    if inner is None:
        inner = container.injector()(func)
        mapping[func] = inner
    return inner


def _try_containers(
    container_list: Sequence[Container],
    func: Callable[..., object],
    *args: object,
    **kwargs: object,
) -> object:
    last_exc: Exception | None = None
    for container in container_list:
        inner = _get_inner(container, func)
        try:
            return inner(*args, **kwargs)
        except (DependencyNotFound, InvalidProvider) as e:
            last_exc = e
            continue
    if last_exc is not None:
        raise last_exc
    raise DependencyNotFound(object)


async def _try_containers_async(
    container_list: Sequence[Container],
    func: Callable[..., object],
    *args: object,
    **kwargs: object,
) -> object:
    last_exc: Exception | None = None
    for container in container_list:
        inner = _get_inner(container, func)
        try:
            inner_async = cast("Callable[..., Awaitable[object]]", inner)
            return await inner_async(*args, **kwargs)
        except (DependencyNotFound, InvalidProvider) as e:
            last_exc = e
            continue
    if last_exc is not None:
        raise last_exc
    raise DependencyNotFound(object)


def _create_wrapper(
    containers: Container | Sequence[Container],
    func: Callable[P, R],
    sig: inspect.Signature,
) -> Callable[..., R]:
    if not isinstance(containers, (list, tuple)):
        container_list = cast("Sequence[Container]", (containers,))
    else:
        container_list = containers

    if inspect.iscoroutinefunction(func):

        async def _awrapper(*args: object, **kwargs: object) -> object:
            return await _try_containers_async(
                container_list, cast("Callable[..., object]", func), *args, **kwargs
            )

        wrapper = _awrapper
    else:

        def _swrapper(*args: object, **kwargs: object) -> object:
            return _try_containers(
                container_list, cast("Callable[..., object]", func), *args, **kwargs
            )

        wrapper = _swrapper

    from contextlib import suppress

    with suppress(Exception):
        setattr(wrapper, "__signature__", sig.replace(parameters=()))
    return cast("Callable[..., R]", wrapper)


def inject(func: Callable[P, R]) -> Callable[..., R]:
    sig = inspect.signature(func)
    containers = _get_container()
    return _create_wrapper(containers, func, sig)


__all__ = ["init", "inject"]



import pytest

from spritze import Container, Depends, Scope, init, inject, provider


class ServiceA:
    pass


class ServiceB:
    def __init__(self, a: ServiceA):
        self.a: ServiceA = a


class TestContainer(Container):
    __test__: bool = False

    def __init__(self):
        self.a_call_count: int = 0
        super().__init__()

    @provider(scope=Scope.APP)
    def a(self) -> ServiceA:
        self.a_call_count += 1
        return ServiceA()

    @provider(scope=Scope.REQUEST)
    def b(self, a: ServiceA) -> ServiceB:
        return ServiceB(a=a)


@pytest.mark.asyncio
async def test_scopes_and_injection():
    container = TestContainer()
    init(container)

    @inject
    async def transaction_1(b1: Depends[ServiceB], a1: Depends[ServiceA]):
        assert isinstance(b1, ServiceB)
        assert isinstance(b1.a, ServiceA)
        assert b1.a is a1
        return a1, b1

    a1, b1 = await transaction_1()
    assert container.a_call_count == 1

    @inject
    async def transaction_2(b2: Depends[ServiceB], a2: Depends[ServiceA]):
        assert isinstance(b2, ServiceB)
        assert isinstance(b2.a, ServiceA)
        assert b2.a is a2
        return a2, b2

    a2, b2 = await transaction_2()
    assert container.a_call_count == 1
    assert a1 is a2
    assert b1 is not b2
    assert b1.a is b2.a


def test_reuse_container():
    c = TestContainer()
    init(c)

    @inject
    def handler(a: Depends[ServiceA]):
        return a

    a1 = handler()
    a2 = handler()
    assert a1 is a2



from spritze import Container, Scope, provider


def test_provider_decorator():
    class D(Container):
        @provider(scope=Scope.APP)
        def foo(self) -> int:
            return 1

    c = D()
    assert c.resolve(int) == 1



# Copyright 2025 Arseny Seliverstov
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Core components
from .application.decorators import provider, singleton, transient
from .application.global_container import init, inject
from .entities.transient import Transient
from .infrastructure.context import context
from .repositories.container_repository import Container
from .value_objects.dependency_marker import DependencyMarker, Depends
from .value_objects.scope import Scope

__all__ = [
    # Core
    "Container",
    "Scope",
    "Transient",
    "Depends",
    "DependencyMarker",
    # Decorators
    "provider",
    "singleton",
    "transient",
    "inject",
    "init",
    # Context
    "context",
]



from __future__ import annotations

from enum import Enum, auto


class ProviderType(Enum):
    SIMPLE = auto()
    ASYNC = auto()
    GEN = auto()
    ASYNC_GEN = auto()


__all__ = ["ProviderType"]
